import click

from anyscale.controllers.cloud_controller import CloudController


@click.group(
    "cloud",
    short_help="Configure cloud provider authentication for Anyscale.",
    help="""Configure cloud provider authentication and setup
to allow Anyscale to launch instances in your account.""",
)
def cloud_cli() -> None:
    pass


@cloud_cli.command(name="delete", help="Delete cloud")
@click.argument("cloud-name", required=False)
@click.option("--cloud-id", help="Cloud id to delete.", required=False)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
def cloud_delete(cloud_name: str, cloud_id: str, yes: bool) -> None:
    CloudController().delete_cloud(
        cloud_name=cloud_name, cloud_id=cloud_id, skip_confirmation=yes
    )


@cloud_cli.command(name="setup", help="Set up a cloud provider.")
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    prompt="Provider",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option("--region", help="Region to set up the credentials in.")
@click.option("--name", help="Name of the cloud.", required=True, prompt="Name")
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
def setup_cloud(provider: str, region: str, name: str, yes: bool) -> None:
    CloudController().setup_cloud(provider=provider, region=region, name=name)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class StopSessionOptions(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'terminate': 'bool',
        'workers_only': 'bool',
        'keep_min_workers': 'bool',
        'delete': 'bool'
    }

    attribute_map = {
        'terminate': 'terminate',
        'workers_only': 'workers_only',
        'keep_min_workers': 'keep_min_workers',
        'delete': 'delete'
    }

    def __init__(self, terminate=None, workers_only=None, keep_min_workers=None, delete=False, local_vars_configuration=None):  # noqa: E501
        """StopSessionOptions - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._terminate = None
        self._workers_only = None
        self._keep_min_workers = None
        self._delete = None
        self.discriminator = None

        self.terminate = terminate
        self.workers_only = workers_only
        self.keep_min_workers = keep_min_workers
        if delete is not None:
            self.delete = delete

    @property
    def terminate(self):
        """Gets the terminate of this StopSessionOptions.  # noqa: E501


        :return: The terminate of this StopSessionOptions.  # noqa: E501
        :rtype: bool
        """
        return self._terminate

    @terminate.setter
    def terminate(self, terminate):
        """Sets the terminate of this StopSessionOptions.


        :param terminate: The terminate of this StopSessionOptions.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and terminate is None:  # noqa: E501
            raise ValueError("Invalid value for `terminate`, must not be `None`")  # noqa: E501

        self._terminate = terminate

    @property
    def workers_only(self):
        """Gets the workers_only of this StopSessionOptions.  # noqa: E501


        :return: The workers_only of this StopSessionOptions.  # noqa: E501
        :rtype: bool
        """
        return self._workers_only

    @workers_only.setter
    def workers_only(self, workers_only):
        """Sets the workers_only of this StopSessionOptions.


        :param workers_only: The workers_only of this StopSessionOptions.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and workers_only is None:  # noqa: E501
            raise ValueError("Invalid value for `workers_only`, must not be `None`")  # noqa: E501

        self._workers_only = workers_only

    @property
    def keep_min_workers(self):
        """Gets the keep_min_workers of this StopSessionOptions.  # noqa: E501


        :return: The keep_min_workers of this StopSessionOptions.  # noqa: E501
        :rtype: bool
        """
        return self._keep_min_workers

    @keep_min_workers.setter
    def keep_min_workers(self, keep_min_workers):
        """Sets the keep_min_workers of this StopSessionOptions.


        :param keep_min_workers: The keep_min_workers of this StopSessionOptions.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and keep_min_workers is None:  # noqa: E501
            raise ValueError("Invalid value for `keep_min_workers`, must not be `None`")  # noqa: E501

        self._keep_min_workers = keep_min_workers

    @property
    def delete(self):
        """Gets the delete of this StopSessionOptions.  # noqa: E501


        :return: The delete of this StopSessionOptions.  # noqa: E501
        :rtype: bool
        """
        return self._delete

    @delete.setter
    def delete(self, delete):
        """Sets the delete of this StopSessionOptions.


        :param delete: The delete of this StopSessionOptions.  # noqa: E501
        :type: bool
        """

        self._delete = delete

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StopSessionOptions):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, StopSessionOptions):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

# flake8: noqa
"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

# import models into model package
from openapi_client.models.aws_credentials import AWSCredentials
from openapi_client.models.anyscale_aws_account import AnyscaleAWSAccount
from openapi_client.models.anyscale_version_response import AnyscaleVersionResponse
from openapi_client.models.anyscaleawsaccount_response import AnyscaleawsaccountResponse
from openapi_client.models.anyscaleversionresponse_response import AnyscaleversionresponseResponse
from openapi_client.models.applied_snapshot import AppliedSnapshot
from openapi_client.models.archived_logs_info import ArchivedLogsInfo
from openapi_client.models.archivedlogsinfo_response import ArchivedlogsinfoResponse
from openapi_client.models.autoscaler_credentials import AutoscalerCredentials
from openapi_client.models.autoscalercredentials_response import AutoscalercredentialsResponse
from openapi_client.models.autosync_session_id import AutosyncSessionId
from openapi_client.models.autosyncsessionid_list_response import AutosyncsessionidListResponse
from openapi_client.models.autosyncsessionid_response import AutosyncsessionidResponse
from openapi_client.models.awscredentials_response import AwscredentialsResponse
from openapi_client.models.change_password_params import ChangePasswordParams
from openapi_client.models.cloud import Cloud
from openapi_client.models.cloud_list_response import CloudListResponse
from openapi_client.models.cloud_name_options import CloudNameOptions
from openapi_client.models.cloud_providers import CloudProviders
from openapi_client.models.cloud_response import CloudResponse
from openapi_client.models.cloud_types import CloudTypes
from openapi_client.models.cluster_config import ClusterConfig
from openapi_client.models.clusterconfig_response import ClusterconfigResponse
from openapi_client.models.create_from_github_options import CreateFromGithubOptions
from openapi_client.models.create_organization_invitation import CreateOrganizationInvitation
from openapi_client.models.create_session_from_snapshot_options import CreateSessionFromSnapshotOptions
from openapi_client.models.create_snapshot_options import CreateSnapshotOptions
from openapi_client.models.create_user import CreateUser
from openapi_client.models.execute_command_options import ExecuteCommandOptions
from openapi_client.models.execute_command_response import ExecuteCommandResponse
from openapi_client.models.execute_shell_command_options import ExecuteShellCommandOptions
from openapi_client.models.executecommandresponse_response import ExecutecommandresponseResponse
from openapi_client.models.feature_flag_response import FeatureFlagResponse
from openapi_client.models.featureflagresponse_response import FeatureflagresponseResponse
from openapi_client.models.gateway_request import GatewayRequest
from openapi_client.models.gatewayrequest_response import GatewayrequestResponse
from openapi_client.models.github_project import GithubProject
from openapi_client.models.githubproject_response import GithubprojectResponse
from openapi_client.models.http_validation_error import HTTPValidationError
from openapi_client.models.head_ip import HeadIp
from openapi_client.models.headip_response import HeadipResponse
from openapi_client.models.json_dumped_gateway_response_message import JsonDumpedGatewayResponseMessage
from openapi_client.models.json_patch_operation import JsonPatchOperation
from openapi_client.models.list_response_metadata import ListResponseMetadata
from openapi_client.models.login_user_params import LoginUserParams
from openapi_client.models.logs_output import LogsOutput
from openapi_client.models.logsoutput_response import LogsoutputResponse
from openapi_client.models.organization import Organization
from openapi_client.models.organization_availability import OrganizationAvailability
from openapi_client.models.organization_invitation import OrganizationInvitation
from openapi_client.models.organization_invitation_base import OrganizationInvitationBase
from openapi_client.models.organization_permission_level import OrganizationPermissionLevel
from openapi_client.models.organization_response import OrganizationResponse
from openapi_client.models.organizationavailability_response import OrganizationavailabilityResponse
from openapi_client.models.organizationinvitation_response import OrganizationinvitationResponse
from openapi_client.models.organizationinvitationbase_response import OrganizationinvitationbaseResponse
from openapi_client.models.permission_level import PermissionLevel
from openapi_client.models.project import Project
from openapi_client.models.project_base import ProjectBase
from openapi_client.models.project_collaborator import ProjectCollaborator
from openapi_client.models.project_collaborator_value import ProjectCollaboratorValue
from openapi_client.models.project_default_session_name import ProjectDefaultSessionName
from openapi_client.models.project_list_response import ProjectListResponse
from openapi_client.models.project_response import ProjectResponse
from openapi_client.models.projectbase_response import ProjectbaseResponse
from openapi_client.models.projectcollaborator_list_response import ProjectcollaboratorListResponse
from openapi_client.models.projectdefaultsessionname_response import ProjectdefaultsessionnameResponse
from openapi_client.models.request_organization_identifiers_params import RequestOrganizationIdentifiersParams
from openapi_client.models.request_password_reset_params import RequestPasswordResetParams
from openapi_client.models.reset_password_params import ResetPasswordParams
from openapi_client.models.s3_download_location import S3DownloadLocation
from openapi_client.models.server_session_token import ServerSessionToken
from openapi_client.models.serversessiontoken_response import ServersessiontokenResponse
from openapi_client.models.session import Session
from openapi_client.models.session_command import SessionCommand
from openapi_client.models.session_command_finish_options import SessionCommandFinishOptions
from openapi_client.models.session_command_id import SessionCommandId
from openapi_client.models.session_describe import SessionDescribe
from openapi_client.models.session_details import SessionDetails
from openapi_client.models.session_finish_up_options import SessionFinishUpOptions
from openapi_client.models.session_history_item import SessionHistoryItem
from openapi_client.models.session_id import SessionId
from openapi_client.models.session_list_response import SessionListResponse
from openapi_client.models.session_overview import SessionOverview
from openapi_client.models.session_report_command_options import SessionReportCommandOptions
from openapi_client.models.session_response import SessionResponse
from openapi_client.models.session_ssh_key import SessionSshKey
from openapi_client.models.session_starting_up_data import SessionStartingUpData
from openapi_client.models.session_state import SessionState
from openapi_client.models.session_state_data import SessionStateData
from openapi_client.models.session_stopping_data import SessionStoppingData
from openapi_client.models.session_up_options import SessionUpOptions
from openapi_client.models.session_up_response import SessionUpResponse
from openapi_client.models.sessioncommand_list_response import SessioncommandListResponse
from openapi_client.models.sessioncommandid_response import SessioncommandidResponse
from openapi_client.models.sessiondescribe_response import SessiondescribeResponse
from openapi_client.models.sessiondetails_response import SessiondetailsResponse
from openapi_client.models.sessionhistoryitem_list_response import SessionhistoryitemListResponse
from openapi_client.models.sessionid_response import SessionidResponse
from openapi_client.models.sessionoverview_list_response import SessionoverviewListResponse
from openapi_client.models.sessionsshkey_response import SessionsshkeyResponse
from openapi_client.models.sessionupresponse_response import SessionupresponseResponse
from openapi_client.models.snapshot import Snapshot
from openapi_client.models.snapshot_create import SnapshotCreate
from openapi_client.models.snapshot_delete import SnapshotDelete
from openapi_client.models.snapshot_files import SnapshotFiles
from openapi_client.models.snapshot_id import SnapshotID
from openapi_client.models.snapshot_list import SnapshotList
from openapi_client.models.snapshot_list_item import SnapshotListItem
from openapi_client.models.snapshot_project_file import SnapshotProjectFile
from openapi_client.models.snapshot_response import SnapshotResponse
from openapi_client.models.snapshotcreate_response import SnapshotcreateResponse
from openapi_client.models.snapshotdelete_response import SnapshotdeleteResponse
from openapi_client.models.snapshotfiles_response import SnapshotfilesResponse
from openapi_client.models.snapshotid_response import SnapshotidResponse
from openapi_client.models.snapshotlist_response import SnapshotlistResponse
from openapi_client.models.start_session_options import StartSessionOptions
from openapi_client.models.stop_session_options import StopSessionOptions
from openapi_client.models.take_snapshot_options import TakeSnapshotOptions
from openapi_client.models.update_organization_collaborator import UpdateOrganizationCollaborator
from openapi_client.models.update_project_collaborator import UpdateProjectCollaborator
from openapi_client.models.upload_session_command_logs_locations import UploadSessionCommandLogsLocations
from openapi_client.models.uploadsessioncommandlogslocations_response import UploadsessioncommandlogslocationsResponse
from openapi_client.models.user_info import UserInfo
from openapi_client.models.user_resend_email_options import UserResendEmailOptions
from openapi_client.models.userinfo_response import UserinfoResponse
from openapi_client.models.validation_error import ValidationError
from openapi_client.models.write_cloud import WriteCloud
from openapi_client.models.write_cluster_config import WriteClusterConfig
from openapi_client.models.write_project import WriteProject
from openapi_client.models.write_project_collaborator import WriteProjectCollaborator
from openapi_client.models.write_project_collaborator_value import WriteProjectCollaboratorValue
from openapi_client.models.write_session import WriteSession
from openapi_client.models.write_snapshot import WriteSnapshot



"""
Fetches data required and formats output for `anyscale cloud` commands.
"""

import json
import os
import time
from typing import Optional

from click import ClickException
from openapi_client.rest import ApiException  # type: ignore

from anyscale.api import get_api_client
from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
from anyscale.client.openapi_client.models.write_cloud import WriteCloud  # type: ignore
from anyscale.cloud import get_cloud_id_and_name
import anyscale.conf
from anyscale.util import (
    _get_role,
    _resource,
    confirm,
    format_api_exception,
    get_available_regions,
    launch_gcp_cloud_setup,
)


class CloudController(object):
    def __init__(self, api_client: DefaultApi = None):
        if api_client is None:
            api_client = get_api_client()
        self.api_client = api_client

    def delete_cloud(
        self,
        cloud_name: Optional[str],
        cloud_id: Optional[str],
        skip_confirmation: bool,
    ) -> bool:
        """
        Deletes a cloud by name or id.
        """

        if not cloud_id and not cloud_name:
            raise ClickException("Must either provide the cloud name or cloud id.")

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )

        confirm(
            f"You'll lose access to existing sessions created with cloud {cloud_id} if you drop it.\nContinue?",
            skip_confirmation,
        )

        with format_api_exception(ApiException):
            resp = self.api_client.delete_cloud_api_v2_clouds_cloud_id_delete(
                cloud_id=cloud_id
            )

        assert resp == {}
        print(f"Deleted cloud {cloud_name}")

        return True

    def setup_cloud(self, provider: str, region: Optional[str], name: str) -> None:
        """
        Sets up a cloud provider
        """

        if provider == "aws":
            # If the region is blank, change it to the default for AWS.
            if region is None:
                region = "us-west-2"
            self.setup_aws(region=region, name=name)
        elif provider == "gcp":
            # If the region is blank, change it to the default for GCP.
            if region is None:
                region = "us-west1"
            # TODO: interactive setup process through the CLI?
            launch_gcp_cloud_setup(name=name, region=region)
        else:
            raise ClickException(
                f"Invalid Cloud provider: {provider}. Available providers are [aws, gcp]."
            )

    def setup_aws(self, region: str, name: str) -> None:
        os.environ["AWS_DEFAULT_REGION"] = region
        regions_available = get_available_regions()
        if region not in regions_available:
            raise ClickException(
                f"Region '{region}' is not available. Regions availables are {regions_available}"
            )

        confirm(
            "\nYou are about to give anyscale full access to EC2 and IAM in your AWS account.\n\n"
            "Continue?",
            False,
        )

        with format_api_exception(ApiException):
            user_info = self.api_client.get_user_info_api_v2_userinfo_get().result

        self.setup_aws_cross_account_role(region, user_info.id, name)

        # Sleep for 5 seconds to make sure the policies take effect.
        time.sleep(5)

        print("AWS credentials setup complete!")
        print(
            "You can revoke the access at any time by deleting anyscale IAM user/role in your account."
        )
        print("Head over to the web UI to create new sessions in your AWS account!")

    def setup_aws_cross_account_role(
        self, region: str, user_id: str, name: str
    ) -> None:
        with format_api_exception(ApiException):
            response = (
                self.api_client.get_anyscale_aws_account_api_v2_clouds_anyscale_aws_account_get()
            )

        anyscale_aws_account = response.result.anyscale_aws_account
        anyscale_aws_iam_role_policy = {
            "Version": "2012-10-17",
            "Statement": {
                "Sid": "1",
                "Effect": "Allow",
                "Principal": {"AWS": anyscale_aws_account},
                "Action": "sts:AssumeRole",
            },
        }

        role = _get_role(anyscale.conf.ANYSCALE_IAM_ROLE_NAME, region)

        if role is None:
            iam = _resource("iam", region)
            iam.create_role(
                RoleName=anyscale.conf.ANYSCALE_IAM_ROLE_NAME,
                AssumeRolePolicyDocument=json.dumps(anyscale_aws_iam_role_policy),
            )
            role = _get_role(anyscale.conf.ANYSCALE_IAM_ROLE_NAME, region)

        assert role is not None, "Failed to create IAM role!"

        role.attach_policy(PolicyArn="arn:aws:iam::aws:policy/AmazonEC2FullAccess")
        role.attach_policy(PolicyArn="arn:aws:iam::aws:policy/IAMFullAccess")

        print(f"Using IAM role {role.arn}")

        with format_api_exception(ApiException):
            self.api_client.create_cloud_api_v2_clouds_post(
                write_cloud=WriteCloud(
                    provider="AWS",
                    region=region,
                    credentials=role.arn,
                    creator_id=user_id,
                    name=name,
                )
            )



from contextlib import contextmanager
import logging
import os
import re
import subprocess
import tempfile
from threading import Event, Thread
import time
from typing import Any, Dict, Iterator, List

from openapi_client.rest import ApiException  # type: ignore
import ray

import anyscale
from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
from anyscale.util import format_api_exception, get_rsync_command, send_json_request

logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
logger = logging.getLogger(__file__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)

if anyscale.conf.AWS_PROFILE is not None:
    logger.info("Using AWS profile %s", anyscale.conf.AWS_PROFILE)
    os.environ["AWS_PROFILE"] = anyscale.conf.AWS_PROFILE

TIME_BETWEEN_HEARTBEATS = 30  # seconds


@contextmanager
def managed_autosync_session(
    session_id: str, api_client: DefaultApi = None
) -> Iterator[str]:
    # Register to the API that we enabled autosync
    with format_api_exception(ApiException):
        resp = api_client.register_autosync_session_api_v2_autosync_sessions_post(
            session_id
        )

    autosync_session_id = resp.result.id
    heartbeat_thread = AutosyncHeartbeat(autosync_session_id)
    heartbeat_thread.start()

    try:
        yield autosync_session_id
    finally:
        heartbeat_thread.finish.set()
        with format_api_exception(ApiException):
            api_client.deregister_autosync_session_api_v2_autosync_sessions_autosync_session_id_delete(
                autosync_session_id
            )
        heartbeat_thread.join()
        print("Autosync finished.")


class AutosyncHeartbeat(Thread):
    def __init__(self, autosync_session_id: str):
        super().__init__()
        self.autosync_session_id = autosync_session_id
        self.finish = Event()

    def run(self) -> None:
        while not self.finish.is_set():
            try:
                heartbeat_status_str = "Last heartbeat sent at {}".format(
                    time.strftime("%x %X %Z")
                )  # Locale date, locale time, timezone string.

                # Keep the heartbeat status output a single line.
                heartbeat_status_str = "\033[1A\033[0K" + heartbeat_status_str
                print(heartbeat_status_str)
                send_json_request(
                    "/api/v2/autosync_sessions/{}/heartbeat".format(
                        self.autosync_session_id
                    ),
                    {},
                    method="POST",
                )
            except Exception as e:
                print("Error sending heartbeat:", e)
            self.finish.wait(TIME_BETWEEN_HEARTBEATS)


def perform_autosync_synchronization(
    fswatch_command: List[str],
    env: Dict[str, Any],
    ssh_command: List[str],
    source: str,
    ssh_user: str,
    head_ip: str,
    target: str,
    sync_git: bool,
) -> None:
    # Check if current directory is a git repo:
    try:
        subprocess.check_call(  # noqa: B1
            ["git", "-C", ".", "rev-parse"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        is_git_repo = True
    except subprocess.CalledProcessError:
        is_git_repo = False

    # Perform synchronization whenever there is a change. We batch together
    # multiple updates and then call rsync on them.
    with subprocess.Popen(
        fswatch_command, stdout=subprocess.PIPE, env=env,
    ) as proc:  # noqa: B1
        while True:
            files = []
            while proc.stdout:
                path = proc.stdout.readline().strip().decode()
                if path == "NoOp":
                    break
                else:
                    relpath = os.path.relpath(path, source)
                    files.append(relpath)
            if files:
                if is_git_repo:
                    # Filter out gitignore files
                    try:
                        ignored_files = subprocess.check_output(  # noqa: B1
                            ["git", "check-ignore", *files]
                        )
                        files_to_ignore = ignored_files.decode().splitlines()
                        for file in files_to_ignore:
                            files.remove(file)
                    except subprocess.CalledProcessError:
                        # Failed check-ignore just means we won't ignore any files
                        # For example, if the directory is not a git repo, this call will fail.
                        pass

                with tempfile.NamedTemporaryFile(mode="w") as modified_files:
                    for f in files:
                        # Avoid rsyncing temporary files that may disappear between
                        # now and when rsync builds a file list.
                        # This matches files ending with: .swp, ~, .tmp or 4913 (neovim)
                        if re.match("(.*(.sw[a-e,g-z]|~|.tmp)$|^4913$)", f):
                            continue
                        modified_files.write(f + "\n")
                    modified_files.flush()
                    command, rsync_env = get_rsync_command(
                        ssh_command,
                        source,
                        ssh_user,
                        head_ip,
                        target,
                        sync_git,
                        modified_files=modified_files,
                    )
                    logger.info("Calling rsync due to detected file update.")
                    logger.debug("Command: {command}".format(command=command))
                    temp_sync = subprocess.run(command, env=rsync_env)  # noqa: B1
                    if temp_sync.returncode != 0:
                        logger.error(
                            f"Secondary sync failed with error code: {temp_sync.returncode}, autosync will continue running."
                        )



import anyscale.client.openapi_client as openapi_client  # type: ignore
from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: ignore
import anyscale.conf
from anyscale.util import load_credentials


def instantiate_api_client(no_cli_token: bool = False) -> DefaultApi:
    if not no_cli_token and anyscale.conf.CLI_TOKEN is None:
        anyscale.conf.CLI_TOKEN = load_credentials()
    configuration = openapi_client.Configuration(host=anyscale.conf.ANYSCALE_HOST)

    if no_cli_token:
        api_client = openapi_client.ApiClient(configuration)
    else:
        api_client = openapi_client.ApiClient(
            configuration, cookie=f"cli_token={anyscale.conf.CLI_TOKEN}"
        )
    api_instance = openapi_client.DefaultApi(api_client)
    return api_instance


def get_api_client() -> DefaultApi:
    if _api_client.api_client is None:
        _api_client.api_client = instantiate_api_client()

    return _api_client.api_client


class _ApiClient(object):
    api_client: DefaultApi = None


_api_client = _ApiClient()



import os
from sys import path

anyscale_dir = os.path.dirname(os.path.abspath(__file__))
path.append(os.path.join(anyscale_dir, "client"))
anyscale_ray_dir = os.path.join(anyscale_dir, "anyscale_ray")
path.insert(0, anyscale_ray_dir)

# Set path before anyscale imports
from anyscale.report import report  # noqa: E402

__version__ = os.getenv("ANYSCALE_CLI_VERSION", "0.3.14")

__all__ = ["report"]

ANYSCALE_ENV = os.environ.copy()
ANYSCALE_ENV["PYTHONPATH"] = anyscale_ray_dir + ":" + ANYSCALE_ENV.get("PYTHONPATH", "")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SessionOverview(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'project_id': 'str',
        'project_name': 'str',
        'session_id': 'str',
        'active': 'bool',
        'session_name': 'str',
        'creator_email': 'str',
        'price_per_h': 'float',
        'price': 'float',
        'total_h': 'float'
    }

    attribute_map = {
        'project_id': 'project_id',
        'project_name': 'project_name',
        'session_id': 'session_id',
        'active': 'active',
        'session_name': 'session_name',
        'creator_email': 'creator_email',
        'price_per_h': 'price_per_h',
        'price': 'price',
        'total_h': 'total_h'
    }

    def __init__(self, project_id=None, project_name=None, session_id=None, active=None, session_name=None, creator_email=None, price_per_h=None, price=None, total_h=None, local_vars_configuration=None):  # noqa: E501
        """SessionOverview - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._project_id = None
        self._project_name = None
        self._session_id = None
        self._active = None
        self._session_name = None
        self._creator_email = None
        self._price_per_h = None
        self._price = None
        self._total_h = None
        self.discriminator = None

        self.project_id = project_id
        self.project_name = project_name
        self.session_id = session_id
        self.active = active
        self.session_name = session_name
        self.creator_email = creator_email
        self.price_per_h = price_per_h
        self.price = price
        self.total_h = total_h

    @property
    def project_id(self):
        """Gets the project_id of this SessionOverview.  # noqa: E501


        :return: The project_id of this SessionOverview.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this SessionOverview.


        :param project_id: The project_id of this SessionOverview.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def project_name(self):
        """Gets the project_name of this SessionOverview.  # noqa: E501


        :return: The project_name of this SessionOverview.  # noqa: E501
        :rtype: str
        """
        return self._project_name

    @project_name.setter
    def project_name(self, project_name):
        """Sets the project_name of this SessionOverview.


        :param project_name: The project_name of this SessionOverview.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_name is None:  # noqa: E501
            raise ValueError("Invalid value for `project_name`, must not be `None`")  # noqa: E501

        self._project_name = project_name

    @property
    def session_id(self):
        """Gets the session_id of this SessionOverview.  # noqa: E501


        :return: The session_id of this SessionOverview.  # noqa: E501
        :rtype: str
        """
        return self._session_id

    @session_id.setter
    def session_id(self, session_id):
        """Sets the session_id of this SessionOverview.


        :param session_id: The session_id of this SessionOverview.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and session_id is None:  # noqa: E501
            raise ValueError("Invalid value for `session_id`, must not be `None`")  # noqa: E501

        self._session_id = session_id

    @property
    def active(self):
        """Gets the active of this SessionOverview.  # noqa: E501


        :return: The active of this SessionOverview.  # noqa: E501
        :rtype: bool
        """
        return self._active

    @active.setter
    def active(self, active):
        """Sets the active of this SessionOverview.


        :param active: The active of this SessionOverview.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and active is None:  # noqa: E501
            raise ValueError("Invalid value for `active`, must not be `None`")  # noqa: E501

        self._active = active

    @property
    def session_name(self):
        """Gets the session_name of this SessionOverview.  # noqa: E501


        :return: The session_name of this SessionOverview.  # noqa: E501
        :rtype: str
        """
        return self._session_name

    @session_name.setter
    def session_name(self, session_name):
        """Sets the session_name of this SessionOverview.


        :param session_name: The session_name of this SessionOverview.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and session_name is None:  # noqa: E501
            raise ValueError("Invalid value for `session_name`, must not be `None`")  # noqa: E501

        self._session_name = session_name

    @property
    def creator_email(self):
        """Gets the creator_email of this SessionOverview.  # noqa: E501


        :return: The creator_email of this SessionOverview.  # noqa: E501
        :rtype: str
        """
        return self._creator_email

    @creator_email.setter
    def creator_email(self, creator_email):
        """Sets the creator_email of this SessionOverview.


        :param creator_email: The creator_email of this SessionOverview.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_email is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_email`, must not be `None`")  # noqa: E501

        self._creator_email = creator_email

    @property
    def price_per_h(self):
        """Gets the price_per_h of this SessionOverview.  # noqa: E501


        :return: The price_per_h of this SessionOverview.  # noqa: E501
        :rtype: float
        """
        return self._price_per_h

    @price_per_h.setter
    def price_per_h(self, price_per_h):
        """Sets the price_per_h of this SessionOverview.


        :param price_per_h: The price_per_h of this SessionOverview.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and price_per_h is None:  # noqa: E501
            raise ValueError("Invalid value for `price_per_h`, must not be `None`")  # noqa: E501

        self._price_per_h = price_per_h

    @property
    def price(self):
        """Gets the price of this SessionOverview.  # noqa: E501


        :return: The price of this SessionOverview.  # noqa: E501
        :rtype: float
        """
        return self._price

    @price.setter
    def price(self, price):
        """Sets the price of this SessionOverview.


        :param price: The price of this SessionOverview.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and price is None:  # noqa: E501
            raise ValueError("Invalid value for `price`, must not be `None`")  # noqa: E501

        self._price = price

    @property
    def total_h(self):
        """Gets the total_h of this SessionOverview.  # noqa: E501


        :return: The total_h of this SessionOverview.  # noqa: E501
        :rtype: float
        """
        return self._total_h

    @total_h.setter
    def total_h(self, total_h):
        """Sets the total_h of this SessionOverview.


        :param total_h: The total_h of this SessionOverview.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and total_h is None:  # noqa: E501
            raise ValueError("Invalid value for `total_h`, must not be `None`")  # noqa: E501

        self._total_h = total_h

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionOverview):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionOverview):
            return True

        return self.to_dict() != other.to_dict()



from unittest.mock import Mock, patch

from click import ClickException
import pytest

from anyscale.client.openapi_client.models import (  # type: ignore
    AnyscaleAWSAccount,
    AnyscaleawsaccountResponse,
    Cloud,
    CloudResponse,
)
from anyscale.controllers.cloud_controller import CloudController


@pytest.fixture()  # type: ignore
def mock_api_client(cloud_test_data: Cloud) -> Mock:
    mock_api_client = Mock()
    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get = Mock(
        return_value=CloudResponse(result=cloud_test_data)
    )
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post = Mock(
        return_value=CloudResponse(result=cloud_test_data)
    )
    mock_api_client.delete_cloud_api_v2_clouds_cloud_id_delete = Mock(return_value={})

    return mock_api_client


def test_setup_cloud_aws(mock_api_client: Mock) -> None:
    with patch.object(
        CloudController, "setup_aws", return_value=None
    ) as mock_setup_aws:
        cloud_controller = CloudController(api_client=mock_api_client)
        cloud_controller.setup_cloud(provider="aws", region=None, name="test-aws")

        mock_setup_aws.assert_called_once_with(region="us-west-2", name="test-aws")


def test_setup_cloud_gcp(mock_api_client: Mock) -> None:
    mock_launch_gcp_cloud_setup = Mock(return_value=None)
    with patch.multiple(
        "anyscale.controllers.cloud_controller",
        launch_gcp_cloud_setup=mock_launch_gcp_cloud_setup,
    ):
        cloud_controller = CloudController(api_client=mock_api_client)
        cloud_controller.setup_cloud(provider="gcp", region=None, name="test-gcp")

        mock_launch_gcp_cloud_setup.assert_called_once_with(
            region="us-west1", name="test-gcp"
        )


def test_setup_cloud_invalid_provider(mock_api_client: Mock) -> None:
    cloud_controller = CloudController(api_client=mock_api_client)
    with pytest.raises(ClickException):
        cloud_controller.setup_cloud(
            provider="azure", region="azure-west-1", name="invalid cloud provider"
        )


def test_delete_cloud_by_name(cloud_test_data: Cloud, mock_api_client: Mock) -> None:
    cloud_controller = CloudController(api_client=mock_api_client)
    success = cloud_controller.delete_cloud(
        cloud_id=None, cloud_name=cloud_test_data.name, skip_confirmation=True
    )
    assert success

    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.assert_called_once_with(
        cloud_name_options={"name": cloud_test_data.name}
    )
    mock_api_client.delete_cloud_api_v2_clouds_cloud_id_delete(
        cloud_id=cloud_test_data.id
    )


def test_delete_cloud_by_id(cloud_test_data: Cloud, mock_api_client: Mock) -> None:
    cloud_controller = CloudController(api_client=mock_api_client)
    success = cloud_controller.delete_cloud(
        cloud_id=cloud_test_data.id, cloud_name=None, skip_confirmation=True
    )
    assert success

    mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
        cloud_id=cloud_test_data.id
    )
    mock_api_client.delete_cloud_api_v2_clouds_cloud_id_delete(
        cloud_id=cloud_test_data.id
    )


def test_delete_cloud_without_name_or_id(mock_api_client: Mock) -> None:
    cloud_controller = CloudController(api_client=mock_api_client)

    with pytest.raises(ClickException):
        cloud_controller.delete_cloud(None, None, True)


def test_setup_cross_region(mock_api_client: Mock) -> None:
    mock_api_client.get_anyscale_aws_account_api_v2_clouds_anyscale_aws_account_get = Mock(
        return_value=AnyscaleawsaccountResponse(
            result=AnyscaleAWSAccount(anyscale_aws_account="aws_account_type")
        )
    )
    mock_role = Mock()
    mock_role.attach_policy = Mock()
    mock_role.arn = "ARN"
    with patch.multiple(
        "anyscale.controllers.cloud_controller", _get_role=Mock(return_value=mock_role)
    ):
        cloud_controller = CloudController(api_client=mock_api_client)
        cloud_controller.setup_aws_cross_account_role("us-west-2", "user_id", "name")
    mock_api_client.get_anyscale_aws_account_api_v2_clouds_anyscale_aws_account_get.assert_called_once()
    mock_role.attach_policy.assert_called()



import subprocess
from subprocess import CalledProcessError
from typing import Any, Dict, List
from unittest.mock import Mock, patch

import pytest

from anyscale.autosync_heartbeat import (
    managed_autosync_session,
    perform_autosync_synchronization,
)


def test_managed_autosync_session() -> None:
    mock_api_client = Mock()
    mock_api_client.register_autosync_session_api_v2_autosync_sessions_post.return_value.result.id = (
        1
    )
    mock_api_client.deregister_autosync_session_api_v2_autosync_sessions_autosync_session_id_delete.return_value = (
        None
    )

    with managed_autosync_session("id", mock_api_client):
        pass

    mock_api_client.register_autosync_session_api_v2_autosync_sessions_post.assert_called_once_with(
        "id"
    )
    mock_api_client.deregister_autosync_session_api_v2_autosync_sessions_autosync_session_id_delete.assert_called_once_with(
        1
    )


# Tests perform_autosync_synchronization for a git repository with nothing ignored
def test_perform_autosync_synchronization() -> None:
    mock_proc = Mock()
    mock_proc.stdout = Mock()

    calls = 0

    def readline(*args: List[Any], **kwargs: Dict[str, Any]) -> bytes:
        nonlocal calls
        if calls == 0:
            calls += 1
            return b"test-file.py"
        else:
            return b"NoOp"

    mock_proc.stdout.readline.side_effect = readline

    with patch(
        "anyscale.autosync_heartbeat.subprocess.check_call"
    ) as mock_check_call, patch(
        "anyscale.autosync_heartbeat.subprocess.Popen"
    ) as mock_popen, patch(
        "anyscale.autosync_heartbeat.subprocess.check_output"
    ) as mock_check_output, patch(
        "anyscale.autosync_heartbeat.subprocess.run"
    ) as mock_run, patch(
        "anyscale.autosync_heartbeat.get_rsync_command"
    ) as mock_get_rsync_command, patch(
        "anyscale.autosync_heartbeat.tempfile.NamedTemporaryFile"
    ) as mock_temp_file:
        mock_check_call.return_value = None
        mock_popen.return_value.__enter__.return_value = mock_proc
        # Git check-ignore returns nothing to ignore
        mock_check_output.return_value = b""
        # Raise exception to exit out of infinite loop
        mock_run.side_effect = Exception()
        mock_get_rsync_command.return_value = (["rsync", "test"], {"test-env": "test"})
        mock_file = Mock()
        mock_temp_file.return_value.__enter__.return_value = mock_file
        mock_file.write.return_value = None

        with pytest.raises(Exception):
            perform_autosync_synchronization(
                ["fswatch"],
                {},
                ["ssh"],
                "project_folder",
                "test-user",
                "1.1.1.1",
                "target_folder",
                False,
            )

        mock_check_call.assert_called_once_with(
            ["git", "-C", ".", "rev-parse"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        mock_popen.assert_called_once_with(["fswatch"], stdout=subprocess.PIPE, env={})

        mock_check_output.assert_called_once_with(
            ["git", "check-ignore", "../test-file.py"]
        )

        mock_file.write.assert_called_once_with("../test-file.py\n")

        mock_get_rsync_command.assert_called_once_with(
            ["ssh"],
            "project_folder",
            "test-user",
            "1.1.1.1",
            "target_folder",
            False,
            modified_files=mock_file,
        )

        mock_run.assert_called_once_with(["rsync", "test"], env={"test-env": "test"})


# Tests perform_autosync_synchronization for a git repository with a file that is ignored
def test_perform_autosync_synchronization_with_ignored_files() -> None:
    mock_proc = Mock()
    mock_proc.stdout = Mock()

    calls = 0

    def readline(*args: List[Any], **kwargs: Dict[str, Any]) -> bytes:
        nonlocal calls
        if calls == 0:
            calls += 1
            return b"test-file.py"
        else:
            return b"NoOp"

    mock_proc.stdout.readline.side_effect = readline

    with patch(
        "anyscale.autosync_heartbeat.subprocess.check_call"
    ) as mock_check_call, patch(
        "anyscale.autosync_heartbeat.subprocess.Popen"
    ) as mock_popen, patch(
        "anyscale.autosync_heartbeat.subprocess.check_output"
    ) as mock_check_output, patch(
        "anyscale.autosync_heartbeat.subprocess.run"
    ) as mock_run, patch(
        "anyscale.autosync_heartbeat.get_rsync_command"
    ) as mock_get_rsync_command, patch(
        "anyscale.autosync_heartbeat.tempfile.NamedTemporaryFile"
    ) as mock_temp_file:
        mock_check_call.return_value = None
        mock_popen.return_value.__enter__.return_value = mock_proc
        # Git check-ignore returns a file to ignore
        mock_check_output.return_value = b"../test-file.py"
        # Raise exception to exit out of infinite loop
        mock_run.side_effect = Exception()
        mock_get_rsync_command.return_value = (["rsync", "test"], {"test-env": "test"})
        mock_file = Mock()
        mock_temp_file.return_value.__enter__.return_value = mock_file
        mock_file.write.return_value = None

        with pytest.raises(Exception):
            perform_autosync_synchronization(
                ["fswatch"],
                {},
                ["ssh"],
                "project_folder",
                "test-user",
                "1.1.1.1",
                "target_folder",
                False,
            )

        mock_check_call.assert_called_once_with(
            ["git", "-C", ".", "rev-parse"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        mock_popen.assert_called_once_with(["fswatch"], stdout=subprocess.PIPE, env={})

        mock_check_output.assert_called_once_with(
            ["git", "check-ignore", "../test-file.py"]
        )

        # No modified files (All ignored)
        mock_file.write.assert_not_called()

        mock_get_rsync_command.assert_called_once_with(
            ["ssh"],
            "project_folder",
            "test-user",
            "1.1.1.1",
            "target_folder",
            False,
            modified_files=mock_file,
        )

        mock_run.assert_called_once_with(["rsync", "test"], env={"test-env": "test"})


# Tests perform_autosync_synchronization works when not in a git repository
def test_perform_autosync_synchronization_non_git_repo() -> None:
    mock_proc = Mock()
    mock_proc.stdout = Mock()

    calls = 0

    def readline(*args: List[Any], **kwargs: Dict[str, Any]) -> bytes:
        nonlocal calls
        if calls == 0:
            calls += 1
            return b"test-file.py"
        else:
            return b"NoOp"

    mock_proc.stdout.readline.side_effect = readline

    with patch(
        "anyscale.autosync_heartbeat.subprocess.check_call"
    ) as mock_check_call, patch(
        "anyscale.autosync_heartbeat.subprocess.Popen"
    ) as mock_popen, patch(
        "anyscale.autosync_heartbeat.subprocess.check_output"
    ) as mock_check_output, patch(
        "anyscale.autosync_heartbeat.subprocess.run"
    ) as mock_run, patch(
        "anyscale.autosync_heartbeat.get_rsync_command"
    ) as mock_get_rsync_command, patch(
        "anyscale.autosync_heartbeat.tempfile.NamedTemporaryFile"
    ) as mock_temp_file:
        mock_check_call.side_effect = CalledProcessError(128, "test-cmd")
        mock_popen.return_value.__enter__.return_value = mock_proc
        mock_check_output.return_value = b""
        # Raise exception to exit out of infinite loop
        mock_run.side_effect = Exception()
        mock_get_rsync_command.return_value = (["rsync", "test"], {"test-env": "test"})

        mock_file = Mock()
        mock_temp_file.return_value.__enter__.return_value = mock_file
        mock_file.write.return_value = None

        with pytest.raises(Exception):
            perform_autosync_synchronization(
                ["fswatch"],
                {},
                ["ssh"],
                "project_folder",
                "test-user",
                "1.1.1.1",
                "target_folder",
                False,
            )

        mock_check_call.assert_called_once_with(
            ["git", "-C", ".", "rev-parse"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        mock_popen.assert_called_once_with(["fswatch"], stdout=subprocess.PIPE, env={})

        # Git check-ignore not called
        mock_check_output.assert_not_called()

        mock_file.write.assert_called_once_with("../test-file.py\n")

        mock_get_rsync_command.assert_called_once_with(
            ["ssh"],
            "project_folder",
            "test-user",
            "1.1.1.1",
            "target_folder",
            False,
            modified_files=mock_file,
        )

        mock_run.assert_called_once_with(["rsync", "test"], env={"test-env": "test"})
